@@ -10,6 +10,7 @@
 #include "hash.h"
 #include "lvm-string.h"
 #include "fs.h"
+#include "defaults.h"
 
 #include <libdevmapper.h>
 #include <limits.h>
@@ -45,7 +46,16 @@ enum {
 	RELOAD = 1,
 	VISIBLE = 2,
 	READWRITE = 3,
-	SUSPENDED = 4
+	SUSPENDED = 4,
+	NOPROPAGATE = 5,
+	TOPLEVEL = 6,
+	REMOVE = 7
+};
+
+enum {
+	MIRR_DISABLED,
+	MIRR_RUNNING,
+	MIRR_COMPLETED
 };
 
 typedef enum {
@@ -73,11 +83,14 @@ struct dev_layer {
 	struct logical_volume *lv;
 
 	/*
-	 * Devices that must be created before this one can be
-	 * created.  Holds str_lists.
+	 * Devices that must be created before this one can be created.
+	 * Reloads get propagated to this list.  Holds str_lists.
 	 */
 	struct list pre_create;
 
+	/* Inverse of pre_create */
+	struct list pre_suspend;
+
 };
 
 struct dl_list {
@@ -85,9 +98,17 @@ struct dl_list {
 	struct dev_layer *dl;
 };
 
+static const char *stripe_filler = NULL;
+static uint32_t mirror_region_size = 0;
+
 struct dev_manager {
 	struct pool *mem;
 
+	struct config_tree *cf;
+	const char *stripe_filler;
+	uint32_t mirror_region_size;
+	uint32_t pvmove_mirror_count;
+
 	char *vg_name;
 
 	/*
@@ -132,7 +153,7 @@ static inline void _clear_flag(struct dev_layer *dl, int bit)
 	dl->flags &= ~(1 << bit);
 }
 
-static int _pre_list_add(struct pool *mem, struct list *pl, char *str)
+static int _pre_list_add(struct pool *mem, struct list *pl, const char *str)
 {
 	struct str_list *sl;
 	struct list *plh;
@@ -142,7 +163,10 @@ static int _pre_list_add(struct pool *mem, struct list *pl, char *str)
 		return 0;
 	}
 
+	/* Already in list? */
 	list_iterate(plh, pl) {
+		if (!strcmp(str, list_item(plh, struct str_list)->str))
+			return 1;
 	}
 
 	if (!(sl = pool_alloc(mem, sizeof(*sl)))) {
@@ -262,7 +286,8 @@ static char *_build_dlid(struct pool *mem, const char *lvid, const char *layer)
 /*
  * Low level device-layer operations.
  */
-static struct dm_task *_setup_task(const char *name, const char *uuid, int task)
+static struct dm_task *_setup_task(const char *name, const char *uuid,
+				   uint32_t *event_nr, int task)
 {
 	struct dm_task *dmt;
 
@@ -277,6 +302,9 @@ static struct dm_task *_setup_task(const char *name, const char *uuid, int task)
 	if (uuid && *uuid)
 		dm_task_set_uuid(dmt, uuid);
 
+	if (event_nr)
+		dm_task_set_event_nr(dmt, *event_nr);
+
 	return dmt;
 }
 
@@ -287,7 +315,7 @@ static int _info_run(const char *name, const char *uuid, struct dm_info *info,
 	struct dm_task *dmt;
 	const char *u;
 
-	if (!(dmt = _setup_task(name, uuid, DM_DEVICE_INFO))) {
+	if (!(dmt = _setup_task(name, uuid, 0, DM_DEVICE_INFO))) {
 		stack;
 		return 0;
 	}
@@ -341,7 +369,7 @@ static int _status_run(const char *name, const char *uuid,
 	char *type = NULL;
 	char *params = NULL;
 
-	if (!(dmt = _setup_task(name, uuid, DM_DEVICE_STATUS))) {
+	if (!(dmt = _setup_task(name, uuid, 0, DM_DEVICE_STATUS))) {
 		stack;
 		return 0;
 	}
@@ -375,6 +403,11 @@ static int _status_run(const char *name, const char *uuid,
 	return r;
 }
 
+static int _status(const char *name, const char *uuid,
+		   unsigned long long *start, unsigned long long *length,
+		   char **type, uint32_t type_size, char **params,
+		   uint32_t param_size) __attribute__ ((unused));
+
 static int _status(const char *name, const char *uuid,
 		   unsigned long long *start, unsigned long long *length,
 		   char **type, uint32_t type_size, char **params,
@@ -392,6 +425,135 @@ static int _status(const char *name, const char *uuid,
 	return 0;
 }
 
+static int _percent_run(struct dev_manager *dm, const char *name,
+			const char *uuid,
+			const char *target_type, int wait,
+			struct logical_volume *lv, float *percent,
+			uint32_t *event_nr)
+{
+	int r = 0;
+	struct dm_task *dmt;
+	struct dm_info info;
+	void *next = NULL;
+	uint64_t start, length;
+	char *type = NULL;
+	char *params = NULL;
+	float percent2;
+	struct list *segh = &lv->segments;
+	struct lv_segment *seg = NULL;
+
+	uint64_t numerator, denominator;
+	uint64_t total_numerator = 0, total_denominator = 0;
+
+	*percent = -1;
+
+	if (!(dmt = _setup_task(name, uuid, event_nr,
+				wait ? DM_DEVICE_WAITEVENT : DM_DEVICE_STATUS)))
+	{
+		stack;
+		return 0;
+	}
+
+	if (!dm_task_run(dmt)) {
+		stack;
+		goto out;
+	}
+
+	if (!dm_task_get_info(dmt, &info) || !info.exists) {
+		stack;
+		goto out;
+	}
+
+	if (event_nr)
+		*event_nr = info.event_nr;
+
+	do {
+		next = dm_get_next_target(dmt, next, &start, &length, &type,
+					  &params);
+		if (lv) {
+			if (!(segh = list_next(&lv->segments, segh))) {
+				log_error("Number of segments in active LV %s "
+					  "does not match metadata", lv->name);
+				goto out;
+			}
+			seg = list_item(segh, struct lv_segment);
+		}
+
+		if (!type || !params || strcmp(type, target_type))
+			continue;
+
+		/* Mirror? */
+		if (!strcmp(type, "mirror")) {
+			log_debug("Mirror status: %s", params);
+			if (sscanf(params, "%*d %*x:%*x %*x:%*x %" PRIu64
+				   "/%" PRIu64, &numerator,
+				   &denominator) != 2) {
+				log_error("Failure parsing mirror status: %s",
+					  params);
+				goto out;
+			}
+			total_numerator += numerator;
+			total_denominator += denominator;
+
+			if (seg && (seg->status & PVMOVE))
+				seg->extents_moved = dm->mirror_region_size *
+				    numerator / lv->vg->extent_size;
+			continue;
+		}
+
+		if (strcmp(type, "snapshot"))
+			continue;
+
+		/* Snapshot */
+		if (index(params, '/')) {
+			if (sscanf(params, "%" PRIu64 "/%" PRIu64,
+				   &numerator, &denominator) == 2) {
+				total_numerator += numerator;
+				total_denominator += denominator;
+			}
+			continue;
+		} else if (sscanf(params, "%f", &percent2) == 1) {
+			*percent += percent2;
+			*percent /= 2;
+		}
+	} while (next);
+
+	if (lv && (segh = list_next(&lv->segments, segh))) {
+		log_error("Number of segments in active LV %s does not "
+			  "match metadata", lv->name);
+		goto out;
+	}
+
+	if (total_denominator)
+		*percent = (float) total_numerator *100 / total_denominator;
+	else
+		*percent = 100;
+
+	log_debug("Mirror percent: %f", *percent);
+	r = 1;
+
+      out:
+	dm_task_destroy(dmt);
+	return r;
+}
+
+static int _percent(struct dev_manager *dm, const char *name, const char *uuid,
+		    const char *target_type, int wait,
+		    struct logical_volume *lv, float *percent,
+		    uint32_t *event_nr)
+{
+	if (uuid && *uuid
+	    && _percent_run(dm, NULL, uuid, target_type, wait, lv, percent,
+			    event_nr))
+		return 1;
+
+	if (name && _percent_run(dm, name, NULL, target_type, wait, lv, percent,
+				 event_nr))
+		return 1;
+
+	return 0;
+}
+
 static int _rename(struct dev_layer *dl, char *newname)
 {
 	int r = 1;
@@ -399,7 +561,7 @@ static int _rename(struct dev_layer *dl, char *newname)
 
 	log_verbose("Renaming %s to %s", dl->name, newname);
 
-	if (!(dmt = _setup_task(dl->name, NULL, DM_DEVICE_RENAME))) {
+	if (!(dmt = _setup_task(dl->name, NULL, 0, DM_DEVICE_RENAME))) {
 		stack;
 		return 0;
 	}
@@ -430,7 +592,7 @@ static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 
 	log_verbose("Loading %s", dl->name);
 	if (!(dmt = _setup_task(task == DM_DEVICE_CREATE ? dl->name : NULL,
-				dl->dlid, task))) {
+				dl->dlid, 0, task))) {
 		stack;
 		return 0;
 	}
@@ -445,8 +607,19 @@ static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 	}
 
 	/*
-	 * Do we want a specific minor number ?
+	 * Do we want a specific device number ?
 	 */
+	if (dl->lv->major >= 0 && _get_flag(dl, VISIBLE)) {
+		if (!dm_task_set_major(dmt, dl->lv->major)) {
+			log_error("Failed to set major number for %s to %d "
+				  "during activation.", dl->name,
+				  dl->lv->major);
+			goto out;
+		} else
+			log_very_verbose("Set major number for %s to %d.",
+					 dl->name, dl->lv->major);
+	}
+
 	if (dl->lv->minor >= 0 && _get_flag(dl, VISIBLE)) {
 		if (!dm_task_set_minor(dmt, dl->lv->minor)) {
 			log_error("Failed to set minor number for %s to %d "
@@ -467,8 +640,14 @@ static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 			log_very_verbose("Activating %s read-only", dl->name);
 	}
 
-	if (!(r = dm_task_run(dmt)))
+	if (!(r = dm_task_run(dmt))) {
 		log_error("Couldn't load device '%s'.", dl->name);
+		if ((dl->lv->minor >= 0 || dl->lv->major >= 0) &&
+		    _get_flag(dl, VISIBLE))
+			log_error("Perhaps the persistent device number "
+				  "%d:%d is already in use?",
+				  dl->lv->major, dl->lv->minor);
+	}
 
 	if (!dm_task_get_info(dmt, &dl->info)) {
 		stack;
@@ -476,6 +655,15 @@ static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 		goto out;
 	}
 
+	if (!dl->info.exists || !dl->info.live_table) {
+		stack;
+		r = 0;
+		goto out;
+	}
+
+	log_very_verbose("Activated %s %s %03u:%03u", dl->name,
+			 dl->dlid, dl->info.major, dl->info.minor);
+
 	if (r && _get_flag(dl, VISIBLE))
 		fs_add_lv(dl->lv, dl->name);
 
@@ -496,7 +684,7 @@ static int _remove(struct dev_layer *dl)
 	else
 		log_very_verbose("Removing %s", dl->name);
 
-	if (!(dmt = _setup_task(dl->name, NULL, DM_DEVICE_REMOVE))) {
+	if (!(dmt = _setup_task(dl->name, NULL, 0, DM_DEVICE_REMOVE))) {
 		stack;
 		return 0;
 	}
@@ -527,7 +715,7 @@ static int _suspend_or_resume(const char *name, action_t suspend)
 	int task = sus ? DM_DEVICE_SUSPEND : DM_DEVICE_RESUME;
 
 	log_very_verbose("%s %s", sus ? "Suspending" : "Resuming", name);
-	if (!(dmt = _setup_task(name, NULL, task))) {
+	if (!(dmt = _setup_task(name, NULL, 0, task))) {
 		stack;
 		return 0;
 	}
@@ -542,7 +730,7 @@ static int _suspend_or_resume(const char *name, action_t suspend)
 
 static int _suspend(struct dev_layer *dl)
 {
-	if (dl->info.suspended)
+	if (!dl->info.exists || dl->info.suspended)
 		return 1;
 
 	if (!_suspend_or_resume(dl->name, SUSPEND)) {
@@ -556,7 +744,7 @@ static int _suspend(struct dev_layer *dl)
 
 static int _resume(struct dev_layer *dl)
 {
-	if (!dl->info.suspended)
+	if (!dl->info.exists || !dl->info.suspended)
 		return 1;
 
 	if (!_suspend_or_resume(dl->name, RESUME)) {
@@ -577,47 +765,106 @@ static int _resume(struct dev_layer *dl)
  * Emit a target for a given segment.
  * FIXME: tidy this function.
  */
-static int _emit_target(struct dm_task *dmt, struct lv_segment *seg)
+static int _emit_target(struct dev_manager *dm, struct dm_task *dmt,
+			struct lv_segment *seg)
 {
 	char params[1024];
 	uint64_t esize = seg->lv->vg->extent_size;
-	uint32_t s, stripes = seg->stripes;
+	uint32_t s, start_area = 0u, areas = seg->area_count;
 	int w = 0, tw = 0;
-	const char *filler = "/dev/ioerror";
 	const char *target = NULL;
+	const char *trailing_space;
+	int mirror_status;
+	struct dev_layer *dl;
+	char devbuf[10];
 
-	if (stripes == 1) {
-		if (!seg->area[0].pv) {
-			target = "error";
-			goto add_target;
-		} else
+	switch (seg->type) {
+	case SEG_SNAPSHOT:
+		log_error("_emit_target: Internal error: Can't handle "
+			  "SEG_SNAPSHOT");
+		return 0;
+		/* Target formats:
+		 *   linear [device offset]+
+		 *   striped #stripes stripe_size [device offset]+
+		 *   mirror  log_type #log_params [log_params]* 
+		 *           #mirrors [device offset]+
+		 */
+	case SEG_STRIPED:
+		if (areas == 1)
 			target = "linear";
-	} else if (stripes > 1) {
-		target = "striped";
-		if ((tw = lvm_snprintf(params, sizeof(params), "%u %u ",
-				       stripes, seg->stripe_size)) < 0)
+		else if (areas > 1) {
+			target = "striped";
+			if ((tw = lvm_snprintf(params, sizeof(params), "%u %u ",
+					       areas, seg->stripe_size)) < 0)
+				goto error;
+			w = tw;
+		} else {
+			log_error("_emit_target: Internal error: SEG_STRIPED "
+				  "with no stripes");
+			return 0;
+		}
+		break;
+	case SEG_MIRRORED:
+		mirror_status = MIRR_RUNNING;
+		if (seg->status & PVMOVE) {
+			if (seg->extents_moved == seg->area_len) {
+				mirror_status = MIRR_COMPLETED;
+				start_area = 1;
+			} else if (dm->pvmove_mirror_count++) {
+				mirror_status = MIRR_DISABLED;
+				areas = 1;
+			}
+		}
+		if (mirror_status != MIRR_RUNNING) {
+			target = "linear";
+			break;
+		}
+		target = "mirror";
+		if ((tw = lvm_snprintf(params, sizeof(params),
+				       "core 1 %u %u ",
+				       dm->mirror_region_size, areas)) < 0)
 			goto error;
 		w = tw;
+		break;
 	}
 
-	for (s = 0; s < stripes; s++, w += tw) {
-		if (!seg->area[s].pv || !seg->area[s].pv->dev)
+	for (s = start_area; s < areas; s++, w += tw) {
+		trailing_space = (areas - s - 1) ? " " : "";
+		if ((seg->area[s].type == AREA_PV &&
+		     (!seg->area[s].u.pv.pv || !seg->area[s].u.pv.pv->dev)) ||
+		    (seg->area[s].type == AREA_LV && !seg->area[s].u.lv.lv))
 			tw = lvm_snprintf(params + w, sizeof(params) - w,
-					  "%s 0%s", filler,
-					  s == (stripes - 1) ? "" : " ");
-		else
+					  "%s 0%s", dm->stripe_filler,
+					  trailing_space);
+		else if (seg->area[s].type == AREA_PV)
 			tw = lvm_snprintf(params + w, sizeof(params) - w,
 					  "%s %" PRIu64 "%s",
-					  dev_name(seg->area[s].pv->dev),
-					  (seg->area[s].pv->pe_start +
-					   (esize * seg->area[s].pe)),
-					  s == (stripes - 1) ? "" : " ");
+					  dev_name(seg->area[s].u.pv.pv->dev),
+					  (seg->area[s].u.pv.pv->pe_start +
+					   (esize * seg->area[s].u.pv.pe)),
+					  trailing_space);
+		else {
+			if (!(dl = hash_lookup(dm->layers,
+					       seg->area[s].u.lv.lv->lvid.s))) {
+				log_error("device layer %s missing from hash",
+					  seg->area[s].u.lv.lv->lvid.s);
+				return 0;
+			}
+			if (!dm_format_dev(devbuf, sizeof(devbuf), dl->info.major, dl->info.minor)) {
+				log_error("Failed to format device number as dm target (%u,%u)",
+					  dl->info.major, dl->info.minor);
+				return 0;
+			}
+			tw = lvm_snprintf(params + w, sizeof(params) - w,
+					  "%s %" PRIu64 "%s", devbuf,
+					  esize * seg->area[s].u.lv.le,
+					  trailing_space);
+		}
 
 		if (tw < 0)
 			goto error;
 	}
 
-      add_target:
 	log_debug("Adding target: %" PRIu64 " %" PRIu64 " %s %s",
 		  esize * seg->le, esize * seg->len, target, params);
 
@@ -630,7 +877,7 @@ static int _emit_target(struct dm_task *dmt, struct lv_segment *seg)
 	return 1;
 
       error:
-	log_error("Insufficient space to write target parameters.");
+	log_error("Insufficient space in params[] to write target parameters.");
 	return 0;
 }
 
@@ -641,9 +888,11 @@ static int _populate_vanilla(struct dev_manager *dm,
 	struct lv_segment *seg;
 	struct logical_volume *lv = dl->lv;
 
+	dm->pvmove_mirror_count = 0u;
+
 	list_iterate(segh, &lv->segments) {
 		seg = list_item(segh, struct lv_segment);
-		if (!_emit_target(dmt, seg)) {
+		if (!_emit_target(dm, dmt, seg)) {
 			log_error("Unable to build table for '%s'", lv->name);
 			return 0;
 		}
@@ -657,13 +906,20 @@ static int _populate_origin(struct dev_manager *dm,
 {
 	char *real;
 	char params[PATH_MAX + 32];
+	struct dev_layer *dlr;
 
-	if (!(real = _build_name(dm->mem, dm->vg_name, dl->lv->name, "real"))) {
+	if (!(real = _build_dlid(dm->mem, dl->lv->lvid.s, "real"))) {
 		stack;
 		return 0;
 	}
 
-	if (lvm_snprintf(params, sizeof(params), "%s/%s", dm_dir(), real) == -1) {
+	if (!(dlr = hash_lookup(dm->layers, real))) {
+		log_error("Couldn't find real device layer %s in hash", real);
+		return 0;
+	}
+
+	if (!dm_format_dev(params, sizeof(params), dlr->info.major,
+			   dlr->info.minor)) {
 		log_error("Couldn't create origin device parameters for '%s'.",
 			  real);
 		return 0;
@@ -671,7 +927,7 @@ static int _populate_origin(struct dev_manager *dm,
 
 	log_debug("Adding target: 0 %" PRIu64 " snapshot-origin %s",
 		  dl->lv->size, params);
-	if (!dm_task_add_target(dmt, __UINT64_C(0), dl->lv->size,
+	if (!dm_task_add_target(dmt, UINT64_C(0), dl->lv->size,
 				"snapshot-origin", params)) {
 		stack;
 		return 0;
@@ -686,25 +942,51 @@ static int _populate_snapshot(struct dev_manager *dm,
 	char *origin, *cow;
 	char params[PATH_MAX * 2 + 32];
 	struct snapshot *s;
+	struct dev_layer *dlo, *dlc;
+	char devbufo[10], devbufc[10];
 
 	if (!(s = find_cow(dl->lv))) {
 		log_error("Couldn't find snapshot for '%s'.", dl->lv->name);
 		return 0;
 	}
 
-	if (!(origin = _build_name(dm->mem, dm->vg_name,
-				   s->origin->name, "real"))) {
+	if (!(origin = _build_dlid(dm->mem, s->origin->lvid.s, "real"))) {
 		stack;
 		return 0;
 	}
 
-	if (!(cow = _build_name(dm->mem, dm->vg_name, s->cow->name, "cow"))) {
+	if (!(cow = _build_dlid(dm->mem, s->cow->lvid.s, "cow"))) {
 		stack;
 		return 0;
 	}
 
-	if (snprintf(params, sizeof(params), "%s/%s %s/%s P %d",
-		     dm_dir(), origin, dm_dir(), cow, s->chunk_size) == -1) {
+	if (!(dlo = hash_lookup(dm->layers, origin))) {
+		log_error("Couldn't find origin device layer %s in hash",
+			  origin);
+		return 0;
+	}
+
+	if (!(dlc = hash_lookup(dm->layers, cow))) {
+		log_error("Couldn't find cow device layer %s in hash", cow);
+		return 0;
+	}
+
+	if (!dm_format_dev(devbufo, sizeof(devbufo), dlo->info.major,
+			   dlo->info.minor)) {
+		log_error("Couldn't create origin device parameters for '%s'.",
+			  s->origin->name);
+		return 0;
+	}
+
+	if (!dm_format_dev(devbufc, sizeof(devbufc), dlc->info.major,
+			   dlc->info.minor)) {
+		log_error("Couldn't create cow device parameters for '%s'.",
+			  s->cow->name);
+		return 0;
+	}
+
+	if (snprintf(params, sizeof(params), "%s %s P %d", 
+		     devbufo, devbufc, s->chunk_size) == -1) {
 		stack;
 		return 0;
 	}
@@ -712,7 +994,7 @@ static int _populate_snapshot(struct dev_manager *dm,
 	log_debug("Adding target: 0 %" PRIu64 " snapshot %s",
 		  s->origin->size, params);
 	if (!dm_task_add_target
-	    (dmt, __UINT64_C(0), s->origin->size, "snapshot", params)) {
+	    (dmt, UINT64_C(0), s->origin->size, "snapshot", params)) {
 		stack;
 		return 0;
 	}
@@ -723,7 +1005,8 @@ static int _populate_snapshot(struct dev_manager *dm,
 /*
  * dev_manager implementation.
  */
-struct dev_manager *dev_manager_create(const char *vg_name)
+struct dev_manager *dev_manager_create(const char *vg_name,
+				       struct config_tree *cf)
 {
 	struct pool *mem;
 	struct dev_manager *dm;
@@ -739,6 +1022,21 @@ struct dev_manager *dev_manager_create(const char *vg_name)
 	}
 
 	dm->mem = mem;
+	dm->cf = cf;
+	if (!stripe_filler) {
+		stripe_filler = find_config_str(cf->root,
+						"activation/missing_stripe_filler",
+						'/', DEFAULT_STRIPE_FILLER);
+	}
+	dm->stripe_filler = stripe_filler;
+
+	if (!mirror_region_size) {
+		mirror_region_size = 2 * find_config_int(cf->root,
+							 "activation/mirror_region_size",
+							 '/',
+							 DEFAULT_MIRROR_REGION_SIZE);
+	}
+	dm->mirror_region_size = mirror_region_size;
 
 	if (!(dm->vg_name = pool_strdup(dm->mem, vg_name))) {
 		stack;
@@ -796,22 +1094,7 @@ int dev_manager_info(struct dev_manager *dm, const struct logical_volume *lv,
 int dev_manager_snapshot_percent(struct dev_manager *dm,
 				 struct logical_volume *lv, float *percent)
 {
-	char *name, *type, *params;
-	unsigned long long start, length;
-
-	/* FIXME: Use #defines - & move allocations into _status_run ? */
-	uint32_t type_size = 32;
-	uint32_t param_size = 32;
-
-	if (!(type = pool_alloc(dm->mem, sizeof(*type) * type_size))) {
-		stack;
-		return 0;
-	}
-
-	if (!(params = pool_alloc(dm->mem, sizeof(*params) * param_size))) {
-		stack;
-		return 0;
-	}
+	char *name;
 
 	/*
 	 * Build a name for the top layer.
@@ -824,23 +1107,45 @@ int dev_manager_snapshot_percent(struct dev_manager *dm,
 	/*
 	 * Try and get some info on this device.
 	 */
-	log_debug("Getting device status for %s", name);
-	if (!(_status(name, lv->lvid.s, &start, &length, &type, type_size,
-		      &params, param_size))) {
+	log_debug("Getting device status percentage for %s", name);
+	if (!(_percent(dm, name, lv->lvid.s, "snapshot", 0, NULL, percent,
+		       NULL))) {
 		stack;
 		return 0;
 	}
 
-	/* FIXME Ensure this is a *snapshot* target with percentage! */
 	/* FIXME pool_free ? */
 
 	/* If the snapshot isn't available, percent will be -1 */
-	*percent = -1;
+	return 1;
+}
 
-	if (!params)
+/* FIXME Merge with snapshot_percent, auto-detecting target type */
+/* FIXME Cope with more than one target */
+int dev_manager_mirror_percent(struct dev_manager *dm,
+			       struct logical_volume *lv, int wait,
+			       float *percent, uint32_t *event_nr)
+{
+	char *name;
+
+	/*
+	 * Build a name for the top layer.
+	 */
+	if (!(name = _build_name(dm->mem, lv->vg->name, lv->name, NULL))) {
+		stack;
 		return 0;
+	}
 
-	return sscanf(params, "%f", percent);
+	/* FIXME pool_free ? */
+
+	log_debug("Getting device mirror status percentage for %s", name);
+	if (!(_percent(dm, name, lv->lvid.s, "mirror", wait, lv, percent,
+		       event_nr))) {
+		stack;
+		return 0;
+	}
+
+	return 1;
 }
 
 static struct dev_layer *_create_dev(struct dev_manager *dm, char *name,
@@ -868,6 +1173,7 @@ static struct dev_layer *_create_dev(struct dev_manager *dm, char *name,
 		dl->dlid = dlid;
 
 	list_init(&dl->pre_create);
+	list_init(&dl->pre_suspend);
 
 	if (!hash_insert(dm->layers, dl->dlid, dl)) {
 		stack;
@@ -932,19 +1238,69 @@ static struct dev_layer *_lookup(struct dev_manager *dm,
 	return dl;
 }
 
-static int _expand_vanilla(struct dev_manager *dm, struct logical_volume *lv)
+static int _expand_vanilla(struct dev_manager *dm, struct logical_volume *lv,
+			   int was_origin)
 {
 	/*
 	 * only one layer.
 	 */
-	struct dev_layer *dl;
+	struct dev_layer *dl, *dlr;
+	struct list *segh;
+	struct lv_segment *seg;
+	uint32_t s;
 
 	if (!(dl = _create_layer(dm, NULL, lv))) {
 		stack;
 		return 0;
 	}
 	dl->populate = _populate_vanilla;
-	_set_flag(dl, VISIBLE);
+	if (lv->status & VISIBLE_LV) {
+		_set_flag(dl, VISIBLE);
+		_set_flag(dl, TOPLEVEL);
+	}
+
+	if (lv->status & PVMOVE)
+		_set_flag(dl, TOPLEVEL);
+
+	/* Add dependencies for any LVs that segments refer to */
+	list_iterate(segh, &lv->segments) {
+		seg = list_item(segh, struct lv_segment);
+		if (seg->type != SEG_STRIPED && seg->type != SEG_MIRRORED)
+			continue;
+		for (s = 0; s < seg->area_count; s++) {
+			if (seg->area[s].type != AREA_LV)
+				continue;
+			if (!_pre_list_add(dm->mem, &dl->pre_create,
+					   _build_dlid(dm->mem,
+						       seg->area[s].u.lv.
+						       lv->lvid.s, NULL))) {
+				stack;
+				return 0;
+			}
+			_set_flag(dl, NOPROPAGATE);
+		}
+	}
+
+	if (!was_origin)
+		return 1;
+
+	/* Deactivating the last snapshot */
+	if (!(dlr = _create_layer(dm, "real", lv))) {
+		stack;
+		return 0;
+	}
+
+	dlr->populate = _populate_vanilla;
+	_clear_flag(dlr, VISIBLE);
+	_clear_flag(dlr, TOPLEVEL);
+	_set_flag(dlr, REMOVE);
+	
+	/* add the dependency on the real device */
+	if (!_pre_list_add(dm->mem, &dl->pre_create,
+			   pool_strdup(dm->mem, dlr->dlid))) {
+		stack;
+		return 0;
+	}
 
 	return 1;
 }
@@ -961,6 +1317,7 @@ static int _expand_origin_real(struct dev_manager *dm,
 	}
 	dl->populate = _populate_vanilla;
 	_clear_flag(dl, VISIBLE);
+	_clear_flag(dl, TOPLEVEL);
 
 	real_dlid = dl->dlid;
 
@@ -970,6 +1327,7 @@ static int _expand_origin_real(struct dev_manager *dm,
 	}
 	dl->populate = _populate_origin;
 	_set_flag(dl, VISIBLE);
+	_set_flag(dl, TOPLEVEL);
 
 	/* add the dependency on the real device */
 	if (!_pre_list_add(dm->mem, &dl->pre_create,
@@ -989,7 +1347,7 @@ static int _expand_origin(struct dev_manager *dm, struct logical_volume *lv)
 
 	/*
 	 * We only need to create an origin layer if one of our
-	 * snapshots is in the active list.
+	 * snapshots is in the active list
 	 */
 	list_iterate(sh, &dm->active_list) {
 		active = list_item(sh, struct lv_list)->lv;
@@ -997,7 +1355,10 @@ static int _expand_origin(struct dev_manager *dm, struct logical_volume *lv)
 			return _expand_origin_real(dm, lv);
 	}
 
-	return _expand_vanilla(dm, lv);
+	/*
+	 * We're deactivating the last snapshot
+	 */
+	return _expand_vanilla(dm, lv, 1);
 }
 
 static int _expand_snapshot(struct dev_manager *dm, struct logical_volume *lv,
@@ -1016,6 +1377,7 @@ static int _expand_snapshot(struct dev_manager *dm, struct logical_volume *lv,
 	}
 	dl->populate = _populate_vanilla;
 	_clear_flag(dl, VISIBLE);
+	_clear_flag(dl, TOPLEVEL);
 
 	cow_dlid = dl->dlid;
 
@@ -1025,21 +1387,28 @@ static int _expand_snapshot(struct dev_manager *dm, struct logical_volume *lv,
 	}
 	dl->populate = _populate_snapshot;
 	_set_flag(dl, VISIBLE);
+	_set_flag(dl, TOPLEVEL);
 
-	/* add the dependency on the real device */
+	/* add the dependency on the cow device */
 	if (!_pre_list_add(dm->mem, &dl->pre_create,
 			   pool_strdup(dm->mem, cow_dlid))) {
 		stack;
 		return 0;
 	}
 
-	/* add the dependency on the org device */
+	/* add the dependency on the real origin device */
 	if (!_pre_list_add(dm->mem, &dl->pre_create,
 			   _build_dlid(dm->mem, s->origin->lvid.s, "real"))) {
 		stack;
 		return 0;
 	}
 
+	/* add the dependency on the visible origin device */
+	if (!_pre_list_add(dm->mem, &dl->pre_suspend, s->origin->lvid.s)) {
+		stack;
+		return 0;
+	}
+
 	return 1;
 }
 
@@ -1059,7 +1428,7 @@ static int _expand_lv(struct dev_manager *dm, struct logical_volume *lv)
 	else if (lv_is_origin(lv))
 		return _expand_origin(dm, lv);
 
-	return _expand_vanilla(dm, lv);
+	return _expand_vanilla(dm, lv, 0);
 }
 
 /*
@@ -1083,7 +1452,7 @@ static int _trace_layer_marks(struct dev_manager *dm, struct dev_layer *dl,
 			      int flag)
 {
 	struct list *sh;
-	char *dlid;
+	const char *dlid;
 	struct dev_layer *dep;
 
 	list_iterate(sh, &dl->pre_create) {
@@ -1097,6 +1466,10 @@ static int _trace_layer_marks(struct dev_manager *dm, struct dev_layer *dl,
 		if (_get_flag(dep, flag))
 			continue;
 
+		/* FIXME Only propagate LV ACTIVE dependencies for now */
+		if ((flag != ACTIVE) && _get_flag(dl, NOPROPAGATE))
+			continue;
+
 		_set_flag(dep, flag);
 
 		if (!_trace_layer_marks(dm, dep, flag)) {
@@ -1156,25 +1529,89 @@ static int _mark_lvs(struct dev_manager *dm, struct list *lvs, int flag)
 	return 1;
 }
 
-static inline int _suspend_parent(struct dev_layer *parent)
+static int _suspend_parents(struct dev_manager *dm, struct dev_layer *dl)
 {
-	return (!parent || !parent->info.exists || _suspend(parent));
+	struct list *sh;
+	struct dev_layer *dep;
+	const char *dlid;
+
+	list_iterate(sh, &dl->pre_suspend) {
+		dlid = list_item(sh, struct str_list)->str;
+
+		if (!(dep = hash_lookup(dm->layers, dlid))) {
+			log_debug("_suspend_parents couldn't find device "
+				  "layer '%s' - skipping.", dlid);
+			continue;
+		}
+
+		if (!strcmp(dep->dlid, dl->dlid)) {
+			log_error("BUG: pre-suspend loop detected (%s)", dlid);
+			return 0;
+		}
+
+		if (!_suspend_parents(dm, dep)) {
+			stack;
+			return 0;
+		}
+
+		if (dep->info.exists & !_suspend(dep)) {
+			stack;
+			return 0;
+		}
+	}
+
+	return 1;
+}
+
+static int _resume_with_deps(struct dev_manager *dm, struct dev_layer *dl)
+{
+	struct list *sh;
+	struct dev_layer *dep;
+	const char *dlid;
+
+	list_iterate(sh, &dl->pre_create) {
+		dlid = list_item(sh, struct str_list)->str;
+
+		if (!(dep = hash_lookup(dm->layers, dlid))) {
+			log_debug("_resume_with_deps couldn't find device "
+				  "layer '%s' - skipping.", dlid);
+			continue;
+		}
+
+		if (!strcmp(dep->dlid, dl->dlid)) {
+			log_error("BUG: pre-create loop detected (%s)", dlid);
+			return 0;
+		}
+
+		if (!_resume_with_deps(dm, dep)) {
+			stack;
+			return 0;
+		}
+	}
+
+	if (dl->info.exists & !_get_flag(dl, SUSPENDED) &&
+	    !_resume(dl)) {
+		stack;
+		return 0;
+	}
+
+	return 1;
 }
 
 /*
  * Recurses through the tree, ensuring that devices are created
  * in correct order.
  */
-static int _create_rec(struct dev_manager *dm, struct dev_layer *dl,
-		       struct dev_layer *parent)
+static int _create_rec(struct dev_manager *dm, struct dev_layer *dl)
 {
 	struct list *sh;
 	struct dev_layer *dep;
-	char *dlid, *newname, *suffix;
+	const char *dlid;
+	char *newname, *suffix;
 
-	/* FIXME Create and use a _suspend_parents() function instead */
 	/* Suspend? */
-	if (_get_flag(dl, SUSPENDED) && (!_suspend_parent || !_suspend(dl))) {
+	if (_get_flag(dl, SUSPENDED) &&
+	    (!_suspend_parents(dm, dl) || !_suspend(dl))) {
 		stack;
 		return 0;
 	}
@@ -1187,17 +1624,12 @@ static int _create_rec(struct dev_manager *dm, struct dev_layer *dl,
 			return 0;
 		}
 
-		if (!_suspend_parent(parent)) {
-			stack;
-			return 0;
-		}
-
 		if (!strcmp(dep->dlid, dl->dlid)) {
 			log_error("BUG: pre-create loop detected (%s)", dlid);
 			return 0;
 		}
 
-		if (!_create_rec(dm, dep, dl)) {
+		if (!_create_rec(dm, dep)) {
 			stack;
 			return 0;
 		}
@@ -1210,7 +1642,7 @@ static int _create_rec(struct dev_manager *dm, struct dev_layer *dl,
 		newname = _build_name(dm->mem, dm->vg_name, dl->lv->name,
 				      suffix);
 		if (strcmp(newname, dl->name)) {
-			if (!_suspend_parent(parent) ||
+			if (!_suspend_parents(dm, dl) ||
 			    !_suspend(dl) || !_rename(dl, newname)) {
 				stack;
 				return 0;
@@ -1220,7 +1652,7 @@ static int _create_rec(struct dev_manager *dm, struct dev_layer *dl,
 
 	/* Create? */
 	if (!dl->info.exists) {
-		if (!_suspend_parent(parent) ||
+		if (!_suspend_parents(dm, dl) ||
 		    !_load(dm, dl, DM_DEVICE_CREATE)) {
 			stack;
 			return 0;
@@ -1230,21 +1662,16 @@ static int _create_rec(struct dev_manager *dm, struct dev_layer *dl,
 
 	/* Reload? */
 	if (_get_flag(dl, RELOAD) &&
-	    (!_suspend_parent(parent) || !_suspend(dl) ||
+	    (!_suspend_parents(dm, dl) || !_suspend(dl) ||
 	     !_load(dm, dl, DM_DEVICE_RELOAD))) {
 		stack;
 		return 0;
 	}
 
-	/* Resume? */
-	if (!_get_flag(dl, SUSPENDED) && (!_suspend_parent || !_resume(dl))) {
-		stack;
-		return 0;
-	}
-
 	return 1;
 }
 
+
 static int _build_all_layers(struct dev_manager *dm, struct volume_group *vg)
 {
 	struct list *lvh;
@@ -1273,6 +1700,9 @@ static int _fill_in_remove_list(struct dev_manager *dm)
 	hash_iterate(hn, dm->layers) {
 		dl = hash_get_data(dm->layers, hn);
 
+		if (_get_flag(dl, REMOVE))
+			_clear_flag(dl, ACTIVE);
+
 		if (!_get_flag(dl, ACTIVE)) {
 			dll = pool_alloc(dm->mem, sizeof(*dll));
 			if (!dll) {
@@ -1288,6 +1718,55 @@ static int _fill_in_remove_list(struct dev_manager *dm)
 	return 1;
 }
 
+static int _populate_pre_suspend_lists(struct dev_manager *dm)
+{
+	struct hash_node *hn;
+	struct dev_layer *dl;
+	struct list *sh;
+	const char *dlid;
+	struct dev_layer *dep;
+
+	hash_iterate(hn, dm->layers) {
+		dl = hash_get_data(dm->layers, hn);
+
+		list_iterate(sh, &dl->pre_suspend) {
+			dlid = list_item(sh, struct str_list)->str;
+
+			if (!(dep = hash_lookup(dm->layers, dlid))) {
+				log_debug("_populate_pre_suspend_lists: "
+					  "Couldn't find device layer '%s' - "
+					  "skipping.", dlid);
+				continue;
+			}
+
+			if (!_pre_list_add(dm->mem, &dep->pre_create,
+					   dl->dlid)) {
+				stack;
+				return 0;
+			}
+		}
+
+		list_iterate(sh, &dl->pre_create) {
+			dlid = list_item(sh, struct str_list)->str;
+
+			if (!(dep = hash_lookup(dm->layers, dlid))) {
+				log_debug("_populate_pre_suspend_lists: "
+					  "Couldn't find device layer '%s' - "
+					  "skipping.", dlid);
+				continue;
+			}
+
+			if (!_pre_list_add(dm->mem, &dep->pre_suspend,
+					   dl->dlid)) {
+				stack;
+				return 0;
+			}
+		}
+	}
+
+	return 1;
+}
+
 /*
  * Layers are removed in a top-down manner.
  */
@@ -1373,14 +1852,32 @@ static int _execute(struct dev_manager *dm, struct volume_group *vg)
 		return 0;
 	}
 
+	if (!_populate_pre_suspend_lists(dm)) {
+		stack;
+		return 0;
+	}
+
 	/*
 	 * Now only top level devices will be unmarked.
 	 */
 	hash_iterate(hn, dm->layers) {
 		dl = hash_get_data(dm->layers, hn);
 
-		if (_get_flag(dl, ACTIVE) && _get_flag(dl, VISIBLE))
-			_create_rec(dm, dl, NULL);
+		if (_get_flag(dl, ACTIVE) && _get_flag(dl, TOPLEVEL))
+			if (!_create_rec(dm, dl)) {
+				stack;
+				return 0;
+			}
+	}
+
+	/* Resume devices */
+	hash_iterate(hn, dm->layers) {
+		dl = hash_get_data(dm->layers, hn);
+
+		if (!_resume_with_deps(dm, dl)) {
+			stack;
+			return 0;
+		}
 	}
 
 	if (!_remove_old_layers(dm)) {
@@ -1434,38 +1931,39 @@ static int _add_existing_layer(struct dev_manager *dm, const char *name)
 
 static int _scan_existing_devices(struct dev_manager *dm)
 {
-	const char *dev_dir = dm_dir();
 
-	int r = 1;
-	const char *name;
-	struct dirent *dirent;
-	DIR *d;
+	int r = 0;
+	struct dm_names *names;
+	unsigned next = 0;
 
-	if (!(d = opendir(dev_dir))) {
-		log_sys_error("opendir", dev_dir);
+	struct dm_task *dmt;
+
+	if (!(dmt = dm_task_create(DM_DEVICE_LIST)))
 		return 0;
-	}
 
-	while ((dirent = readdir(d))) {
-		name = dirent->d_name;
+	if (!dm_task_run(dmt))
+		goto out;
 
-		if (name[0] == '.')
-			continue;
+	if (!(names = dm_task_get_names(dmt)))
+		goto out;
 
-		/*
-		 * Does this layer belong to us ?
-		 */
-		if (_belong_to_vg(dm->vg_name, name) &&
-		    !_add_existing_layer(dm, name)) {
+	r = 1;
+	if (!names->dev)
+		goto out;
+
+	do {
+		names = (void *) names + next;
+		if (_belong_to_vg(dm->vg_name, names->name) &&
+		    !_add_existing_layer(dm, names->name)) {
 			stack;
 			r = 0;
 			break;
 		}
-	}
-
-	if (closedir(d))
-		log_sys_error("closedir", dev_dir);
+		next = names->next;
+	} while (next);
 
+      out:
+	dm_task_destroy(dmt);
 	return r;
 }
 
@@ -1550,12 +2048,34 @@ static int _remove_lvs(struct dev_manager *dm, struct logical_volume *lv)
 	return _add_lvs(dm->mem, &dm->reload_list, old_origin);
 }
 
+static int _remove_suspended_lvs(struct dev_manager *dm,
+				 struct logical_volume *lv)
+{
+	struct logical_volume *suspended;
+	struct snapshot *s;
+	struct list *sh, *suspend_head;
+
+	suspend_head = &dm->suspend_list;
+
+	/* Remove from list any snapshots with given origin */
+	list_iterate(sh, suspend_head) {
+		suspended = list_item(sh, struct lv_list)->lv;
+		if ((s = find_cow(suspended)) && s->origin == lv) {
+			_remove_lv(suspend_head, suspended);
+		}
+	}
+
+	_remove_lv(suspend_head, lv);
+
+	return 1;
+}
+
 static int _fill_in_active_list(struct dev_manager *dm, struct volume_group *vg)
 {
-	int found;
 	char *dlid;
 	struct list *lvh;
 	struct logical_volume *lv;
+	struct dev_layer *dl;
 
 	list_iterate(lvh, &vg->lvs) {
 		lv = list_item(lvh, struct lv_list)->lv;
@@ -1565,16 +2085,24 @@ static int _fill_in_active_list(struct dev_manager *dm, struct volume_group *vg)
 			return 0;
 		}
 
-		found = hash_lookup(dm->layers, dlid) ? 1 : 0;
+		dl = hash_lookup(dm->layers, dlid);
 		pool_free(dm->mem, dlid);
 
-		if (found) {
-			log_debug("Found active lv %s", lv->name);
+		if (dl) {
+			log_debug("Found active lv %s%s", lv->name,
+				  dl->info.suspended ? " (suspended)" : "");
 
 			if (!_add_lv(dm->mem, &dm->active_list, lv)) {
 				stack;
 				return 0;
 			}
+
+			if (dl->info.suspended) {
+				if (!_add_lv(dm->mem, &dm->suspend_list, lv)) {
+					stack;
+					return 0;
+				}
+			}
 		}
 	}
 
@@ -1610,6 +2138,13 @@ static int _action(struct dev_manager *dm, struct logical_volume *lv,
 		}
 	}
 
+	if (action == SUSPEND || action == RESUME || action == ACTIVATE)
+		/* Get into known state - remove from suspend list if present */
+		if (!_remove_suspended_lvs(dm, lv)) {
+			stack;
+			return 0;
+		}
+
 	if (action == SUSPEND) {
 		if (!_add_lvs(dm->mem, &dm->suspend_list, lv)) {
 			stack;
@@ -1639,3 +2174,8 @@ int dev_manager_suspend(struct dev_manager *dm, struct logical_volume *lv)
 {
 	return _action(dm, lv, SUSPEND);
 }
+
+void dev_manager_exit(void)
+{
+	dm_lib_exit();
+}
