@@ -15,7 +15,6 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <ctype.h>
-#include <asm/page.h>
 
 enum {
 	TOK_INT,
@@ -50,7 +49,7 @@ struct cs {
 	char *filename;
 };
 
-static void _get_token(struct parser *p);
+static void _get_token(struct parser *p, int tok_prev);
 static void _eat_space(struct parser *p);
 static struct config_node *_file(struct parser *p);
 static struct config_node *_section(struct parser *p);
@@ -111,14 +110,15 @@ void destroy_config_tree(struct config_tree *cf)
 	pool_destroy(((struct cs *) cf)->mem);
 }
 
-int read_config_fd(struct config_tree *cf, int fd, const char *file,
-		   off_t offset, uint32_t size, off_t offset2, uint32_t size2,
+int read_config_fd(struct config_tree *cf, struct device *dev,
+		   off_t offset, size_t size, off_t offset2, size_t size2,
 		   checksum_fn_t checksum_fn, uint32_t checksum)
 {
 	struct cs *c = (struct cs *) cf;
 	struct parser *p;
-	off_t mmap_offset;
 	int r = 0;
+	int use_mmap = 1;
+	off_t mmap_offset = 0;
 
 	if (!(p = pool_alloc(c->mem, sizeof(*p)))) {
 		stack;
@@ -126,47 +126,43 @@ int read_config_fd(struct config_tree *cf, int fd, const char *file,
 	}
 	p->mem = c->mem;
 
-	if (size2) {
-		/* FIXME Attempt adjacent mmaps MAP_FIXED into malloced space 
-		 * one PAGE_SIZE larger than required...
-		 */
+	/* Only use mmap with regular files */
+	if (!(dev->flags & DEV_REGULAR) || size2)
+		use_mmap = 0;
+
+	if (use_mmap) {
+		mmap_offset = offset % getpagesize();
+		/* memory map the file */
+		p->fb = mmap((caddr_t) 0, size + mmap_offset, PROT_READ,
+			     MAP_PRIVATE, dev_fd(dev), offset - mmap_offset);
+		if (p->fb == (caddr_t) (-1)) {
+			log_sys_error("mmap", dev_name(dev));
+			goto out;
+		}
+		p->fb = p->fb + mmap_offset;
+	} else {
 		if (!(p->fb = dbg_malloc(size + size2))) {
 			stack;
 			return 0;
 		}
-		if (lseek(fd, offset, SEEK_SET) < 0) {
-			log_sys_error("lseek", file);
+		if (!dev_read(dev, (uint64_t) offset, size, p->fb)) {
+			log_error("Read from %s failed", dev_name(dev));
 			goto out;
 		}
-		if (raw_read(fd, p->fb, size) != size) {
-			log_error("Circular read from %s failed", file);
-			goto out;
+		if (size2) {
+			if (!dev_read(dev, (uint64_t) offset2, size2,
+				      p->fb + size)) {
+				log_error("Circular read from %s failed",
+					  dev_name(dev));
+				goto out;
+			}
 		}
-		if (lseek(fd, offset2, SEEK_SET) < 0) {
-			log_sys_error("lseek", file);
-			goto out;
-		}
-		if (raw_read(fd, p->fb + size, size2) != size2) {
-			log_error("Circular read from %s failed", file);
-			goto out;
-		}
-	} else {
-		mmap_offset = offset % PAGE_SIZE;
-		/* memory map the file */
-		p->fb = mmap((caddr_t) 0, size + mmap_offset, PROT_READ,
-			     MAP_PRIVATE, fd, offset - mmap_offset);
-		if (p->fb == (caddr_t) (-1)) {
-			log_sys_error("mmap", file);
-			goto out;
-		}
-
-		p->fb = p->fb + mmap_offset;
 	}
 
 	if (checksum_fn && checksum !=
 	    (checksum_fn(checksum_fn(INITIAL_CRC, p->fb, size),
 			 p->fb + size, size2))) {
-		log_error("%s: Checksum error", file);
+		log_error("%s: Checksum error", dev_name(dev));
 		goto out;
 	}
 
@@ -175,7 +171,7 @@ int read_config_fd(struct config_tree *cf, int fd, const char *file,
 	/* parse */
 	p->tb = p->te = p->fb;
 	p->line = 1;
-	_get_token(p);
+	_get_token(p, TOK_SECTION_E);
 	if (!(cf->root = _file(p))) {
 		stack;
 		goto out;
@@ -184,12 +180,12 @@ int read_config_fd(struct config_tree *cf, int fd, const char *file,
 	r = 1;
 
       out:
-	if (size2)
+	if (!use_mmap)
 		dbg_free(p->fb);
 	else {
 		/* unmap the file */
-		if (munmap((char *) (p->fb - mmap_offset), size)) {
-			log_sys_error("munmap", file);
+		if (munmap((char *) (p->fb - mmap_offset), size + mmap_offset)) {
+			log_sys_error("munmap", dev_name(dev));
 			r = 0;
 		}
 	}
@@ -201,7 +197,8 @@ int read_config_file(struct config_tree *cf, const char *file)
 {
 	struct cs *c = (struct cs *) cf;
 	struct stat info;
-	int r = 1, fd;
+	struct device *dev;
+	int r = 1;
 
 	if (stat(file, &info)) {
 		log_sys_error("stat", file);
@@ -218,15 +215,20 @@ int read_config_file(struct config_tree *cf, const char *file)
 		return 1;
 	}
 
-	if ((fd = open(file, O_RDONLY)) < 0) {
-		log_sys_error("open", file);
+	if (!(dev = dev_create_file(file, NULL, NULL))) {
+		stack;
 		return 0;
 	}
 
-	r = read_config_fd(cf, fd, file, 0, info.st_size, 0, 0,
+	if (!dev_open_flags(dev, O_RDONLY, 0, 0)) {
+		stack;
+		return 0;
+	}
+
+	r = read_config_fd(cf, dev, 0, (size_t) info.st_size, 0, 0,
 			   (checksum_fn_t) NULL, 0);
 
-	close(fd);
+	dev_close(dev);
 
 	c->timestamp = info.st_mtime;
 	c->filename = pool_strdup(c->mem, file);
@@ -234,6 +236,13 @@ int read_config_file(struct config_tree *cf, const char *file)
 	return r;
 }
 
+time_t config_file_timestamp(struct config_tree *cf)
+{
+	struct cs *c = (struct cs *) cf;
+
+	return c->timestamp;
+}
+
 /*
  * Returns 1 if config file reloaded
  */
@@ -243,7 +252,11 @@ int reload_config_file(struct config_tree **cf)
 	struct cs *c = (struct cs *) *cf;
 	struct cs *new_cs;
 	struct stat info;
-	int r, fd;
+	struct device *dev;
+	int r;
+
+	if (!c->filename)
+		return 0;
 
 	if (stat(c->filename, &info) == -1) {
 		if (errno == ENOENT)
@@ -270,19 +283,25 @@ int reload_config_file(struct config_tree **cf)
 		return 0;
 	}
 
-	if ((fd = open(c->filename, O_RDONLY)) < 0) {
-		log_sys_error("open", c->filename);
-		return 0;
-	}
-
 	if (!(new_cf = create_config_tree())) {
 		log_error("Allocation of new config_tree failed");
 		return 0;
 	}
 
-	r = read_config_fd(new_cf, fd, c->filename, 0, info.st_size, 0, 0,
-			   (checksum_fn_t) NULL, 0);
-	close(fd);
+	if (!(dev = dev_create_file(c->filename, NULL, NULL))) {
+		stack;
+		return 0;
+	}
+
+	if (!dev_open_flags(dev, O_RDONLY, 0, 0)) {
+		stack;
+		return 0;
+	}
+
+	r = read_config_fd(new_cf, dev, 0, (size_t) info.st_size,
+			   0, 0, (checksum_fn_t) NULL, 0);
+
+	dev_close(dev);
 
 	if (r) {
 		new_cs = (struct cs *) new_cf;
@@ -483,7 +502,7 @@ static struct config_value *_value(struct parser *p)
 
 static struct config_value *_type(struct parser *p)
 {
-	/* [0-9]+ | [0-9]*\.[0-9]* | ".*" */
+	/* [+-]{0,1}[0-9]+ | [0-9]*\.[0-9]* | ".*" */
 	struct config_value *v = _create_value(p);
 
 	if (!v)
@@ -526,15 +545,17 @@ static int _match_aux(struct parser *p, int t)
 	if (p->t != t)
 		return 0;
 
-	_get_token(p);
+	_get_token(p, t);
 	return 1;
 }
 
 /*
  * tokeniser
  */
-static void _get_token(struct parser *p)
+static void _get_token(struct parser *p, int tok_prev)
 {
+	int values_allowed = 0;
+
 	p->tb = p->te;
 	_eat_space(p);
 	if (p->tb == p->fe || !*p->tb) {
@@ -542,6 +563,11 @@ static void _get_token(struct parser *p)
 		return;
 	}
 
+	/* Should next token be interpreted as value instead of identifier? */
+	if (tok_prev == TOK_EQ || tok_prev == TOK_ARRAY_B ||
+	    tok_prev == TOK_COMMA)
+		values_allowed = 1;
+
 	p->t = TOK_INT;		/* fudge so the fall through for
 				   floats works */
 	switch (*p->te) {
@@ -589,6 +615,16 @@ static void _get_token(struct parser *p)
 			p->te++;
 		break;
 
+	case '\'':
+		p->t = TOK_STRING;
+		p->te++;
+		while ((p->te != p->fe) && (*p->te) && (*p->te != '\''))
+			p->te++;
+
+		if ((p->te != p->fe) && (*p->te))
+			p->te++;
+		break;
+
 	case '.':
 		p->t = TOK_FLOAT;
 	case '0':
@@ -601,22 +637,27 @@ static void _get_token(struct parser *p)
 	case '7':
 	case '8':
 	case '9':
-		p->te++;
-		while ((p->te != p->fe) && (*p->te)) {
-			if (*p->te == '.') {
-				if (p->t == TOK_FLOAT)
-					break;
-				p->t = TOK_FLOAT;
-			} else if (!isdigit((int) *p->te))
-				break;
+	case '+':
+	case '-':
+		if (values_allowed) {
 			p->te++;
+			while ((p->te != p->fe) && (*p->te)) {
+				if (*p->te == '.') {
+					if (p->t == TOK_FLOAT)
+						break;
+					p->t = TOK_FLOAT;
+				} else if (!isdigit((int) *p->te))
+					break;
+				p->te++;
+			}
+			break;
 		}
-		break;
 
 	default:
 		p->t = TOK_IDENTIFIER;
 		while ((p->te != p->fe) && (*p->te) && !isspace(*p->te) &&
-		       (*p->te != '#') && (*p->te != '='))
+		       (*p->te != '#') && (*p->te != '=') && (*p->te != '{') &&
+		       (*p->te != '}'))
 			p->te++;
 		break;
 	}
@@ -665,7 +706,7 @@ static struct config_node *_create_node(struct parser *p)
 
 static char *_dup_tok(struct parser *p)
 {
-	int len = p->te - p->tb;
+	size_t len = p->te - p->tb;
 	char *str = pool_alloc(p->mem, len + 1);
 	if (!str) {
 		stack;
@@ -680,7 +721,7 @@ static char *_dup_tok(struct parser *p)
  * utility functions
  */
 struct config_node *find_config_node(struct config_node *cn,
-				     const char *path, char sep)
+				     const char *path, const int sep)
 {
 	const char *e;
 
@@ -712,7 +753,7 @@ struct config_node *find_config_node(struct config_node *cn,
 }
 
 const char *find_config_str(struct config_node *cn,
-			    const char *path, char sep, const char *fail)
+			    const char *path, const int sep, const char *fail)
 {
 	struct config_node *n = find_config_node(cn, path, sep);
 
@@ -729,7 +770,7 @@ const char *find_config_str(struct config_node *cn,
 }
 
 int find_config_int(struct config_node *cn, const char *path,
-		    char sep, int fail)
+		    const int sep, int fail)
 {
 	struct config_node *n = find_config_node(cn, path, sep);
 
@@ -744,7 +785,7 @@ int find_config_int(struct config_node *cn, const char *path,
 }
 
 float find_config_float(struct config_node *cn, const char *path,
-			char sep, float fail)
+			const int sep, float fail)
 {
 	struct config_node *n = find_config_node(cn, path, sep);
 
@@ -787,7 +828,7 @@ static int _str_to_bool(const char *str, int fail)
 }
 
 int find_config_bool(struct config_node *cn, const char *path,
-		     char sep, int fail)
+		     const int sep, int fail)
 {
 	struct config_node *n = find_config_node(cn, path, sep);
 	struct config_value *v;
@@ -809,7 +850,7 @@ int find_config_bool(struct config_node *cn, const char *path,
 }
 
 int get_config_uint32(struct config_node *cn, const char *path,
-		      char sep, uint32_t *result)
+		      const int sep, uint32_t *result)
 {
 	struct config_node *n;
 
@@ -823,7 +864,7 @@ int get_config_uint32(struct config_node *cn, const char *path,
 }
 
 int get_config_uint64(struct config_node *cn, const char *path,
-		      char sep, uint64_t *result)
+		      const int sep, uint64_t *result)
 {
 	struct config_node *n;
 
@@ -838,7 +879,7 @@ int get_config_uint64(struct config_node *cn, const char *path,
 }
 
 int get_config_str(struct config_node *cn, const char *path,
-		   char sep, char **result)
+		   const int sep, char **result)
 {
 	struct config_node *n;
 
