@@ -42,7 +42,7 @@ static struct {
 } _segtypes[] = {
 	{
 	SEG_STRIPED, "striped"}, {
-	SEG_MIRROR, "mirror"}, {
+	SEG_MIRRORED, "mirror"}, {
 	SEG_SNAPSHOT, "snapshot"}
 };
 
@@ -70,18 +70,20 @@ uint64_t units_to_bytes(const char *units, char *unit_type)
 	switch (*units) {
 	case 'h':
 	case 'H':
-		v = __UINT64_C(1);
+		v = UINT64_C(1);
 		*unit_type = *units;
 		break;
 	case 's':
 		v *= SECTOR_SIZE;
+		break;
 	case 'b':
 	case 'B':
-		v *= __UINT64_C(1);
+		v *= UINT64_C(1);
 		break;
-#define KILO __UINT64_C(1024)
+#define KILO UINT64_C(1024)
 	case 'k':
 		v *= KILO;
+		break;
 	case 'm':
 		v *= KILO * KILO;
 		break;
@@ -92,7 +94,7 @@ uint64_t units_to_bytes(const char *units, char *unit_type)
 		v *= KILO * KILO * KILO * KILO;
 		break;
 #undef KILO
-#define KILO __UINT64_C(1000)
+#define KILO UINT64_C(1000)
 	case 'K':
 		v *= KILO;
 		break;
@@ -166,8 +168,8 @@ const char *display_size(struct cmd_context *cmd, uint64_t size, size_len_t sl)
 {
 	int s;
 	int suffix = 1;
-	uint64_t byte;
-	uint64_t units = __UINT64_C(1024);
+	uint64_t byte = UINT64_C(0);
+	uint64_t units = UINT64_C(1024);
 	char *size_buf = NULL;
 	const char *size_str[][3] = {
 		{" Terabyte", " TB", "T"},
@@ -193,20 +195,20 @@ const char *display_size(struct cmd_context *cmd, uint64_t size, size_len_t sl)
 		    *size_str[s][2])
 			break;
 
-	if (size == __UINT64_C(0)) {
+	if (size == UINT64_C(0)) {
 		sprintf(size_buf, "0%s", suffix ? size_str[s][sl] : "");
 		return size_buf;
 	}
 
 	if (s < 8) {
 		byte = cmd->current_settings.unit_factor;
-		size *= __UINT64_C(1024);
+		size *= UINT64_C(1024);
 	} else {
 		suffix = 1;
 		if (cmd->current_settings.unit_type == 'H')
-			units = __UINT64_C(1000);
+			units = UINT64_C(1000);
 		else
-			units = __UINT64_C(1024);
+			units = UINT64_C(1024);
 		byte = units * units * units;
 		s = 0;
 		while (size_str[s] && size < byte)
@@ -388,6 +390,9 @@ int lvdisplay_full(struct cmd_context *cmd, struct logical_volume *lv,
 			snap = list_item(slh, struct snapshot_list)->snapshot;
 			snap_active = lv_snapshot_percent(snap->cow,
 							  &snap_percent);
+			if (!snap_active || snap_percent < 0 ||
+			    snap_percent >= 100)
+				snap_active = 0;
 			log_print("                       %s%s/%s [%s]",
 				  lv->vg->cmd->dev_dir, lv->vg->name,
 				  snap->cow->name,
@@ -396,6 +401,8 @@ int lvdisplay_full(struct cmd_context *cmd, struct logical_volume *lv,
 		snap = NULL;
 	} else if ((snap = find_cow(lv))) {
 		snap_active = lv_snapshot_percent(lv, &snap_percent);
+		if (!snap_active || snap_percent < 0 || snap_percent >= 100)
+			snap_active = 0;
 		log_print("LV snapshot status     %s destination for %s%s/%s",
 			  (snap_active > 0) ? "active" : "INACTIVE",
 			  lv->vg->cmd->dev_dir, lv->vg->name,
@@ -461,8 +468,11 @@ int lvdisplay_full(struct cmd_context *cmd, struct logical_volume *lv,
 	log_print("Allocation             %s", get_alloc_string(lv->alloc));
 	log_print("Read ahead sectors     %u", lv->read_ahead);
 
-	if (lv->status & FIXED_MINOR)
+	if (lv->status & FIXED_MINOR) {
+		if (lv->major >= 0)
+			log_print("Persistent major       %d", lv->major);
 		log_print("Persistent minor       %d", lv->minor);
+	}
 
 	if (inkernel)
 		log_print("Block device           %d:%d", info.major,
@@ -475,14 +485,28 @@ int lvdisplay_full(struct cmd_context *cmd, struct logical_volume *lv,
 
 static void _display_stripe(struct lv_segment *seg, uint32_t s, const char *pre)
 {
-	uint32_t len = seg->len / seg->stripes;
+	switch (seg->area[s].type) {
+	case AREA_PV:
+		log_print("%sPhysical volume\t%s", pre,
+			  seg->area[s].u.pv.pv ?
+			  dev_name(seg->area[s].u.pv.pv->dev) : "Missing");
 
-	log_print("%sPhysical volume\t%s", pre,
-		  seg->area[s].pv ? dev_name(seg->area[s].pv->dev) : "Missing");
+		if (seg->area[s].u.pv.pv)
+			log_print("%sPhysical extents\t%d to %d", pre,
+				  seg->area[s].u.pv.pe,
+				  seg->area[s].u.pv.pe + seg->area_len - 1);
+		break;
+	case AREA_LV:
+		log_print("%sLogical volume\t%s", pre,
+			  seg->area[s].u.lv.lv ?
+			  seg->area[s].u.lv.lv->name : "Missing");
 
-	if (seg->area[s].pv)
-		log_print("%sPhysical extents\t%d to %d", pre,
-			  seg->area[s].pe, seg->area[s].pe + len - 1);
+		if (seg->area[s].u.lv.lv)
+			log_print("%sLogical extents\t%d to %d", pre,
+				  seg->area[s].u.lv.le,
+				  seg->area[s].u.lv.le + seg->area_len - 1);
+
+	}
 }
 
 int lvdisplay_segments(struct logical_volume *lv)
@@ -499,7 +523,7 @@ int lvdisplay_segments(struct logical_volume *lv)
 		log_print("Logical extent %u to %u:",
 			  seg->le, seg->le + seg->len - 1);
 
-		if (seg->type == SEG_STRIPED && seg->stripes == 1)
+		if (seg->type == SEG_STRIPED && seg->area_count == 1)
 			log_print("  Type\t\tlinear");
 		else
 			log_print("  Type\t\t%s",
@@ -507,14 +531,14 @@ int lvdisplay_segments(struct logical_volume *lv)
 
 		switch (seg->type) {
 		case SEG_STRIPED:
-			if (seg->stripes == 1)
+			if (seg->area_count == 1)
 				_display_stripe(seg, 0, "  ");
 			else {
-				log_print("  Stripes\t\t%u", seg->stripes);
+				log_print("  Stripes\t\t%u", seg->area_count);
 				log_print("  Stripe size\t\t%u KB",
 					  seg->stripe_size / 2);
 
-				for (s = 0; s < seg->stripes; s++) {
+				for (s = 0; s < seg->area_count; s++) {
 					log_print("  Stripe %d:", s);
 					_display_stripe(seg, s, "    ");
 				}
@@ -522,8 +546,16 @@ int lvdisplay_segments(struct logical_volume *lv)
 			log_print(" ");
 			break;
 		case SEG_SNAPSHOT:
-		case SEG_MIRROR:
-			;
+			break;
+		case SEG_MIRRORED:
+			log_print("  Mirrors\t\t%u", seg->area_count);
+			log_print("  Mirror size\t\t%u", seg->area_len);
+			log_print("  Mirror original:");
+			_display_stripe(seg, 0, "    ");
+			log_print("  Mirror destination:");
+			_display_stripe(seg, 1, "    ");
+			log_print(" ");
+			break;
 		}
 	}
 
