@@ -9,11 +9,9 @@
 #include "pool.h"
 #include "xlate.h"
 #include "filter.h"
-#include "cache.h"
+#include "lvmcache.h"
 
-#include <sys/stat.h>
 #include <fcntl.h>
-#include <linux/kdev_t.h>
 
 #define fail do {stack; return 0;} while(0)
 #define xx16(v) disk->v = xlate16(disk->v)
@@ -132,7 +130,7 @@ static int _munge_formats(struct pv_disk *pvd)
 
 static int _read_pvd(struct device *dev, struct pv_disk *pvd)
 {
-	if (dev_read(dev, __UINT64_C(0), sizeof(*pvd), pvd) != sizeof(*pvd)) {
+	if (!dev_read(dev, UINT64_C(0), sizeof(*pvd), pvd)) {
 		log_very_verbose("Failed to read PV data from %s",
 				 dev_name(dev));
 		return 0;
@@ -157,7 +155,7 @@ static int _read_pvd(struct device *dev, struct pv_disk *pvd)
 
 static int _read_lvd(struct device *dev, uint64_t pos, struct lv_disk *disk)
 {
-	if (dev_read(dev, pos, sizeof(*disk), disk) != sizeof(*disk))
+	if (!dev_read(dev, pos, sizeof(*disk), disk))
 		fail;
 
 	_xlate_lvd(disk);
@@ -169,7 +167,7 @@ static int _read_vgd(struct disk_list *data)
 {
 	struct vg_disk *vgd = &data->vgd;
 	uint64_t pos = data->pvd.vg_on_disk.base;
-	if (dev_read(data->dev, pos, sizeof(*vgd), vgd) != sizeof(*vgd))
+	if (!dev_read(data->dev, pos, sizeof(*vgd), vgd))
 		fail;
 
 	_xlate_vgd(vgd);
@@ -186,8 +184,7 @@ static int _read_uuids(struct disk_list *data)
 	uint64_t end = pos + data->pvd.pv_uuidlist_on_disk.size;
 
 	while (pos < end && num_read < data->vgd.pv_cur) {
-		if (dev_read(data->dev, pos, sizeof(buffer), buffer) !=
-		    sizeof(buffer))
+		if (!dev_read(data->dev, pos, sizeof(buffer), buffer))
 			fail;
 
 		if (!(ul = pool_alloc(data->mem, sizeof(*ul))))
@@ -246,7 +243,7 @@ static int _read_extents(struct disk_list *data)
 	if (!extents)
 		fail;
 
-	if (dev_read(data->dev, pos, len, extents) != len)
+	if (!dev_read(data->dev, pos, len, extents))
 		fail;
 
 	_xlate_extents(extents, data->pvd.pe_total);
@@ -258,21 +255,22 @@ static int _read_extents(struct disk_list *data)
 /* 
  * If exported, remove "PV_EXP" from end of VG name 
  */
-static void _munge_exported_vg(struct disk_list *data)
+void munge_exported_vg(struct pv_disk *pvd)
 {
 	int l;
 	size_t s;
 
-	/* Return if PV not in a VG or VG not exported */
-	if ((!*data->pvd.vg_name) || !(data->vgd.vg_status & VG_EXPORTED))
+	/* Return if PV not in a VG */
+	if ((!*pvd->vg_name))
 		return;
+	/* FIXME also check vgd->status & VG_EXPORTED? */
 
-	l = strlen(data->pvd.vg_name);
+	l = strlen(pvd->vg_name);
 	s = sizeof(EXPORTED_TAG);
-	if (!strncmp(data->pvd.vg_name + l - s + 1, EXPORTED_TAG, s))
-		data->pvd.vg_name[l - s + 1] = '\0';
+	if (!strncmp(pvd->vg_name + l - s + 1, EXPORTED_TAG, s))
+		pvd->vg_name[l - s + 1] = '\0';
 
-	data->pvd.pv_status |= VG_EXPORTED;
+	pvd->pv_status |= VG_EXPORTED;
 }
 
 static struct disk_list *__read_disk(const struct format_type *fmt,
@@ -281,7 +279,7 @@ static struct disk_list *__read_disk(const struct format_type *fmt,
 {
 	struct disk_list *dl = pool_alloc(mem, sizeof(*dl));
 	const char *name = dev_name(dev);
-	struct cache_info *info;
+	struct lvmcache_info *info;
 
 	if (!dl) {
 		stack;
@@ -298,8 +296,11 @@ static struct disk_list *__read_disk(const struct format_type *fmt,
 		goto bad;
 	}
 
-	if (!(info = cache_add(fmt->labeller, dl->pvd.pv_uuid, dev,
-			       dl->pvd.vg_name, NULL)))
+	/* If VG is exported, set VG name back to the real name */
+	munge_exported_vg(&dl->pvd);
+
+	if (!(info = lvmcache_add(fmt->labeller, dl->pvd.pv_uuid, dev,
+				  dl->pvd.vg_name, NULL)))
 		stack;
 	else {
 		info->device_size = xlate32(dl->pvd.pv_size) << SECTOR_SHIFT;
@@ -324,9 +325,6 @@ static struct disk_list *__read_disk(const struct format_type *fmt,
 		goto bad;
 	}
 
-	/* If VG is exported, set VG name back to the real name */
-	_munge_exported_vg(dl);
-
 	/* Update VG cache with what we found */
 	/* vgcache_add(dl->pvd.vg_name, dl->vgd.vg_uuid, dev, fmt); */
 
@@ -367,7 +365,7 @@ struct disk_list *read_disk(const struct format_type *fmt, struct device *dev,
 {
 	struct disk_list *r;
 
-	if (!dev_open(dev, O_RDONLY)) {
+	if (!dev_open(dev)) {
 		stack;
 		return NULL;
 	}
@@ -417,13 +415,13 @@ int read_pvs_in_vg(const struct format_type *fmt, const char *vg_name,
 	struct device *dev;
 	struct disk_list *data = NULL;
 	struct list *vgih;
-	struct cache_vginfo *vginfo;
+	struct lvmcache_vginfo *vginfo;
 
 	/* Fast path if we already saw this VG and cached the list of PVs */
 	if (vg_name && (vginfo = vginfo_from_vgname(vg_name)) &&
 	    vginfo->infos.n) {
 		list_iterate(vgih, &vginfo->infos) {
-			dev = list_item(vgih, struct cache_info)->dev;
+			dev = list_item(vgih, struct lvmcache_info)->dev;
 			if (dev && !(data = read_disk(fmt, dev, mem, vg_name)))
 				break;
 			_add_pv_to_list(head, data);
@@ -465,7 +463,7 @@ static int _write_vgd(struct disk_list *data)
 	uint64_t pos = data->pvd.vg_on_disk.base;
 
 	_xlate_vgd(vgd);
-	if (dev_write(data->dev, pos, sizeof(*vgd), vgd) != sizeof(*vgd))
+	if (!dev_write(data->dev, pos, sizeof(*vgd), vgd))
 		fail;
 
 	_xlate_vgd(vgd);
@@ -488,7 +486,7 @@ static int _write_uuids(struct disk_list *data)
 		}
 
 		ul = list_item(uh, struct uuid_list);
-		if (dev_write(data->dev, pos, NAME_LEN, ul->uuid) != NAME_LEN)
+		if (!dev_write(data->dev, pos, NAME_LEN, ul->uuid))
 			fail;
 
 		pos += NAME_LEN;
@@ -500,7 +498,7 @@ static int _write_uuids(struct disk_list *data)
 static int _write_lvd(struct device *dev, uint64_t pos, struct lv_disk *disk)
 {
 	_xlate_lvd(disk);
-	if (dev_write(dev, pos, sizeof(*disk), disk) != sizeof(*disk))
+	if (!dev_write(dev, pos, sizeof(*disk), disk))
 		fail;
 
 	_xlate_lvd(disk);
@@ -544,7 +542,7 @@ static int _write_extents(struct disk_list *data)
 	uint64_t pos = data->pvd.pe_on_disk.base;
 
 	_xlate_extents(extents, data->pvd.pe_total);
-	if (dev_write(data->dev, pos, len, extents) != len)
+	if (!dev_write(data->dev, pos, len, extents))
 		fail;
 
 	_xlate_extents(extents, data->pvd.pe_total);
@@ -576,7 +574,7 @@ static int _write_pvd(struct disk_list *data)
 	memcpy(buf, &data->pvd, sizeof(struct pv_disk));
 
 	_xlate_pvd((struct pv_disk *) buf);
-	if (dev_write(data->dev, pos, size, buf) != size) {
+	if (!dev_write(data->dev, pos, size, buf)) {
 		dbg_free(buf);
 		fail;
 	}
@@ -642,7 +640,7 @@ static int _write_all_pvd(const struct format_type *fmt, struct disk_list *data)
 {
 	int r;
 
-	if (!dev_open(data->dev, O_WRONLY)) {
+	if (!dev_open(data->dev)) {
 		stack;
 		return 0;
 	}
