@@ -6,6 +6,7 @@
 
 #include "lib.h"
 #include "metadata.h"
+#include "locking.h"
 #include "pv_map.h"
 #include "lvm-string.h"
 #include "toolcontext.h"
@@ -19,9 +20,12 @@ static void _get_extents(struct lv_segment *seg)
 	unsigned int s, count;
 	struct physical_volume *pv;
 
-	for (s = 0; s < seg->stripes; s++) {
-		pv = seg->area[s].pv;
-		count = seg->len / seg->stripes;
+	for (s = 0; s < seg->area_count; s++) {
+		if (seg->area[s].type != AREA_PV)
+			continue;
+
+		pv = seg->area[s].u.pv.pv;
+		count = seg->area_len;
 		pv->pe_alloc_count += count;
 	}
 }
@@ -31,12 +35,17 @@ static void _put_extents(struct lv_segment *seg)
 	unsigned int s, count;
 	struct physical_volume *pv;
 
-	for (s = 0; s < seg->stripes; s++) {
-		pv = seg->area[s].pv;
-		count = seg->len / seg->stripes;
+	for (s = 0; s < seg->area_count; s++) {
+		if (seg->area[s].type != AREA_PV)
+			continue;
 
-		assert(pv->pe_alloc_count >= count);
-		pv->pe_alloc_count -= count;
+		pv = seg->area[s].u.pv.pv;
+
+		if (pv) {
+			count = seg->area_len;
+			assert(pv->pe_alloc_count >= count);
+			pv->pe_alloc_count -= count;
+		}
 	}
 }
 
@@ -58,13 +67,13 @@ static int _alloc_stripe_area(struct logical_volume *lv, uint32_t stripes,
 			      struct pv_area **areas, uint32_t *ix)
 {
 	uint32_t count = lv->le_count - *ix;
-	uint32_t per_area = count / stripes;
+	uint32_t area_len = count / stripes;
 	uint32_t smallest = areas[stripes - 1]->count;
 	uint32_t s;
 	struct lv_segment *seg;
 
-	if (smallest < per_area)
-		per_area = smallest;
+	if (smallest < area_len)
+		area_len = smallest;
 
 	if (!(seg = _alloc_segment(lv->vg->cmd->mem, stripes))) {
 		log_err("Couldn't allocate new stripe segment.");
@@ -74,15 +83,17 @@ static int _alloc_stripe_area(struct logical_volume *lv, uint32_t stripes,
 	seg->lv = lv;
 	seg->type = SEG_STRIPED;
 	seg->le = *ix;
-	seg->len = per_area * stripes;
-	seg->stripes = stripes;
+	seg->len = area_len * stripes;
+	seg->area_len = area_len;
+	seg->area_count = stripes;
 	seg->stripe_size = stripe_size;
 
 	for (s = 0; s < stripes; s++) {
 		struct pv_area *pva = areas[s];
-		seg->area[s].pv = pva->map->pv;
-		seg->area[s].pe = pva->start;
-		consume_pv_area(pva, per_area);
+		seg->area[s].type = AREA_PV;
+		seg->area[s].u.pv.pv = pva->map->pvl->pv;
+		seg->area[s].u.pv.pe = pva->start;
+		consume_pv_area(pva, area_len);
 	}
 
 	list_add(&lv->segments, &seg->list);
@@ -186,10 +197,12 @@ static int _alloc_linear_area(struct logical_volume *lv, uint32_t *ix,
 	seg->type = SEG_STRIPED;
 	seg->le = *ix;
 	seg->len = count;
+	seg->area_len = count;
 	seg->stripe_size = 0;
-	seg->stripes = 1;
-	seg->area[0].pv = map->pv;
-	seg->area[0].pe = pva->start;
+	seg->area_count = 1;
+	seg->area[0].type = AREA_PV;
+	seg->area[0].u.pv.pv = map->pvl->pv;
+	seg->area[0].u.pv.pe = pva->start;
 
 	list_add(&lv->segments, &seg->list);
 
@@ -198,6 +211,47 @@ static int _alloc_linear_area(struct logical_volume *lv, uint32_t *ix,
 	return 1;
 }
 
+static int _alloc_mirrored_area(struct logical_volume *lv, uint32_t *ix,
+				struct pv_map *map, struct pv_area *pva,
+				struct physical_volume *mirrored_pv,
+				uint32_t mirrored_pe)
+{
+	uint32_t count, remaining;
+	struct lv_segment *seg;
+
+	count = pva->count;
+	remaining = lv->le_count - *ix;
+	if (count > remaining)
+		count = remaining;
+
+	if (!(seg = _alloc_segment(lv->vg->cmd->mem, 2))) {
+		log_err("Couldn't allocate new mirrored segment.");
+		return 0;
+	}
+
+	seg->lv = lv;
+	seg->type = SEG_MIRRORED;
+	seg->status = 0u;
+	seg->le = *ix;
+	seg->len = count;
+	seg->area_len = count;
+	seg->stripe_size = 0;
+	seg->area_count = 2;
+	seg->extents_moved = 0u;
+	/* FIXME Remove AREA_PV restriction here? */
+	seg->area[0].type = AREA_PV;
+	seg->area[0].u.pv.pv = mirrored_pv;
+	seg->area[0].u.pv.pe = mirrored_pe;
+	seg->area[1].type = AREA_PV;
+	seg->area[1].u.pv.pv = map->pvl->pv;
+	seg->area[1].u.pv.pe = pva->start;
+	list_add(&lv->segments, &seg->list);
+
+	consume_pv_area(pva, count);
+	*ix += count;
+	return 1;
+}
+
 /*
  * Only one area per pv is allowed, so we search
  * for the biggest area, or the first area that
@@ -222,14 +276,15 @@ static int _alloc_contiguous(struct logical_volume *lv,
 
 		/* first item in the list is the biggest */
 		pva = list_item(pvm->areas.n, struct pv_area);
+		if (pva->count < lv->le_count)
+			continue;
 
 		if (!_alloc_linear_area(lv, &allocated, pvm, pva)) {
 			stack;
 			return 0;
 		}
 
-		if (allocated == lv->le_count)
-			break;
+		break;
 	}
 
 	if (allocated != lv->le_count) {
@@ -242,6 +297,50 @@ static int _alloc_contiguous(struct logical_volume *lv,
 	return 1;
 }
 
+/* FIXME Contiguous depends on *segment* (i.e. stripe) not LV */
+static int _alloc_mirrored(struct logical_volume *lv,
+			   struct list *pvms, uint32_t allocated,
+			   struct physical_volume *mirrored_pv,
+			   uint32_t mirrored_pe)
+{
+	struct list *tmp1;
+	struct pv_map *pvm;
+	struct pv_area *pva;
+	uint32_t max_found = 0;
+
+	/* Try each PV in turn */
+	list_iterate(tmp1, pvms) {
+		pvm = list_item(tmp1, struct pv_map);
+
+		if (list_empty(&pvm->areas))
+			continue;
+
+		/* first item in the list is the biggest */
+		pva = list_item(pvm->areas.n, struct pv_area);
+		if (pva->count < lv->le_count - allocated) {
+			max_found = pva->count;
+			continue;
+		}
+
+		if (!_alloc_mirrored_area(lv, &allocated, pvm, pva,
+					  mirrored_pv, mirrored_pe)) {
+			stack;
+			return 0;
+		}
+
+		break;
+	}
+
+	if (allocated != lv->le_count) {
+		log_error("Insufficient contiguous allocatable extents (%u) "
+			  "for logical volume %s: %u required",
+			  allocated + max_found, lv->name, lv->le_count);
+		return 0;
+	}
+
+	return 1;
+}
+
 /*
  * Areas just get allocated in order until the lv
  * is full.
@@ -279,12 +378,15 @@ static int _alloc_next_free(struct logical_volume *lv,
  * Chooses a correct allocation policy.
  */
 static int _allocate(struct volume_group *vg, struct logical_volume *lv,
-		     struct list *acceptable_pvs, uint32_t allocated,
-		     uint32_t stripes, uint32_t stripe_size)
+		     struct list *allocatable_pvs, uint32_t allocated,
+		     uint32_t stripes, uint32_t stripe_size,
+		     struct physical_volume *mirrored_pv, uint32_t mirrored_pe,
+		     uint32_t status)
 {
 	int r = 0;
 	struct pool *scratch;
 	struct list *pvms, *old_tail = lv->segments.p, *segh;
+	struct lv_segment *seg;
 
 	if (!(scratch = pool_create(1024))) {
 		stack;
@@ -294,12 +396,15 @@ static int _allocate(struct volume_group *vg, struct logical_volume *lv,
 	/*
 	 * Build the sets of available areas on the pv's.
 	 */
-	if (!(pvms = create_pv_maps(scratch, vg, acceptable_pvs)))
+	if (!(pvms = create_pv_maps(scratch, vg, allocatable_pvs)))
 		goto out;
 
 	if (stripes > 1)
 		r = _alloc_striped(lv, pvms, allocated, stripes, stripe_size);
 
+	else if (mirrored_pv)
+		r = _alloc_mirrored(lv, pvms, allocated, mirrored_pv,
+				    mirrored_pe);
 	else if (lv->alloc == ALLOC_CONTIGUOUS)
 		r = _alloc_contiguous(lv, pvms, allocated);
 
@@ -319,8 +424,11 @@ static int _allocate(struct volume_group *vg, struct logical_volume *lv,
 		 * Iterate through the new segments, updating pe
 		 * counts in pv's.
 		 */
-		for (segh = lv->segments.p; segh != old_tail; segh = segh->p)
-			_get_extents(list_item(segh, struct lv_segment));
+		for (segh = lv->segments.p; segh != old_tail; segh = segh->p) {
+			seg = list_item(segh, struct lv_segment);
+			_get_extents(seg);
+			seg->status = status;
+		}
 	} else {
 		/*
 		 * Put the segment list back how we found it.
@@ -334,7 +442,7 @@ static int _allocate(struct volume_group *vg, struct logical_volume *lv,
 	return r;
 }
 
-static char *_generate_lv_name(struct volume_group *vg,
+static char *_generate_lv_name(struct volume_group *vg, const char *format,
 			       char *buffer, size_t len)
 {
 	struct list *lvh;
@@ -344,19 +452,86 @@ static char *_generate_lv_name(struct volume_group *vg,
 	list_iterate(lvh, &vg->lvs) {
 		lv = (list_item(lvh, struct lv_list)->lv);
 
-		if (sscanf(lv->name, "lvol%d", &i) != 1)
+		if (sscanf(lv->name, format, &i) != 1)
 			continue;
 
 		if (i > high)
 			high = i;
 	}
 
-	if (lvm_snprintf(buffer, len, "lvol%d", high + 1) < 0)
+	if (lvm_snprintf(buffer, len, format, high + 1) < 0)
 		return NULL;
 
 	return buffer;
 }
 
+struct logical_volume *lv_create_empty(struct format_instance *fi,
+				       const char *name,
+				       const char *name_format,
+				       uint32_t status,
+				       alloc_policy_t alloc,
+				       struct volume_group *vg)
+{
+	struct cmd_context *cmd = vg->cmd;
+	struct lv_list *ll = NULL;
+	struct logical_volume *lv;
+	char dname[32];
+
+	if (vg->max_lv == vg->lv_count) {
+		log_error("Maximum number of logical volumes (%u) reached "
+			  "in volume group %s", vg->max_lv, vg->name);
+		return NULL;
+	}
+
+	if (!name && !(name = _generate_lv_name(vg, name_format, dname,
+						sizeof(dname)))) {
+		log_error("Failed to generate unique name for the new "
+			  "logical volume");
+		return NULL;
+	}
+
+	log_verbose("Creating logical volume %s", name);
+
+	if (!(ll = pool_zalloc(cmd->mem, sizeof(*ll))) ||
+	    !(ll->lv = pool_zalloc(cmd->mem, sizeof(*ll->lv)))) {
+		log_error("lv_list allocation failed");
+		if (ll)
+			pool_free(cmd->mem, ll);
+		return NULL;
+	}
+
+	lv = ll->lv;
+	lv->vg = vg;
+
+	if (!(lv->name = pool_strdup(cmd->mem, name))) {
+		log_error("lv name strdup failed");
+		if (ll)
+			pool_free(cmd->mem, ll);
+		return NULL;
+	}
+
+	lv->status = status;
+	lv->alloc = alloc;
+	lv->read_ahead = 0;
+	lv->major = -1;
+	lv->minor = -1;
+	lv->size = UINT64_C(0);
+	lv->le_count = 0;
+	list_init(&lv->segments);
+
+	if (fi->fmt->ops->lv_setup && !fi->fmt->ops->lv_setup(fi, lv)) {
+		stack;
+		if (ll)
+			pool_free(cmd->mem, ll);
+		return NULL;
+	}
+
+	vg->lv_count++;
+	list_add(&vg->lvs, &ll->list);
+
+	return lv;
+}
+
 struct logical_volume *lv_create(struct format_instance *fi,
 				 const char *name,
 				 uint32_t status,
@@ -365,12 +540,9 @@ struct logical_volume *lv_create(struct format_instance *fi,
 				 uint32_t stripe_size,
 				 uint32_t extents,
 				 struct volume_group *vg,
-				 struct list *acceptable_pvs)
+				 struct list *allocatable_pvs)
 {
-	struct cmd_context *cmd = vg->cmd;
-	struct lv_list *ll = NULL;
 	struct logical_volume *lv;
-	char dname[32];
 
 	if (!extents) {
 		log_error("Unable to create logical volume %s with no extents",
@@ -384,70 +556,33 @@ struct logical_volume *lv_create(struct format_instance *fi,
 		return NULL;
 	}
 
-	if (vg->max_lv == vg->lv_count) {
-		log_error("Maximum number of logical volumes (%u) reached "
-			  "in volume group %s", vg->max_lv, vg->name);
-		return NULL;
-	}
-
-	if (stripes > list_size(acceptable_pvs)) {
+	if (stripes > list_size(allocatable_pvs)) {
 		log_error("Number of stripes (%u) must not exceed "
 			  "number of physical volumes (%d)", stripes,
-			  list_size(acceptable_pvs));
+			  list_size(allocatable_pvs));
 		return NULL;
 	}
 
-	if (!name && !(name = _generate_lv_name(vg, dname, sizeof(dname)))) {
-		log_error("Failed to generate unique name for the new "
-			  "logical volume");
-		return NULL;
-	}
-
-	log_verbose("Creating logical volume %s", name);
-
-	if (!(ll = pool_zalloc(cmd->mem, sizeof(*ll))) ||
-	    !(ll->lv = pool_zalloc(cmd->mem, sizeof(*ll->lv)))) {
+	if (!(lv = lv_create_empty(fi, name, "lvol%d", status, alloc, vg))) {
 		stack;
 		return NULL;
 	}
 
-	lv = ll->lv;
-
-	lv->vg = vg;
-
-	if (!(lv->name = pool_strdup(cmd->mem, name))) {
-		stack;
-		goto bad;
-	}
-
-	lv->status = status;
-	lv->alloc = alloc;
-	lv->read_ahead = 0;
-	lv->minor = -1;
 	lv->size = (uint64_t) extents *vg->extent_size;
 	lv->le_count = extents;
-	list_init(&lv->segments);
 
-	if (!_allocate(vg, lv, acceptable_pvs, 0u, stripes, stripe_size)) {
+	if (!_allocate(vg, lv, allocatable_pvs, 0u, stripes, stripe_size,
+		       NULL, 0u, 0u)) {
 		stack;
-		goto bad;
+		return NULL;
 	}
 
 	if (fi->fmt->ops->lv_setup && !fi->fmt->ops->lv_setup(fi, lv)) {
 		stack;
-		goto bad;
+		return NULL;
 	}
 
-	vg->lv_count++;
-	list_add(&vg->lvs, &ll->list);
-
 	return lv;
-
-      bad:
-	if (ll)
-		pool_free(cmd->mem, ll);
-
-	return NULL;
 }
 
 int lv_reduce(struct format_instance *fi,
@@ -490,7 +625,7 @@ int lv_reduce(struct format_instance *fi,
 int lv_extend(struct format_instance *fi,
 	      struct logical_volume *lv,
 	      uint32_t stripes, uint32_t stripe_size,
-	      uint32_t extents, struct list *acceptable_pvs)
+	      uint32_t extents, struct list *allocatable_pvs)
 {
 	uint32_t old_le_count = lv->le_count;
 	uint64_t old_size = lv->size;
@@ -498,10 +633,11 @@ int lv_extend(struct format_instance *fi,
 	lv->le_count += extents;
 	lv->size += (uint64_t) extents *lv->vg->extent_size;
 
-	if (!_allocate(lv->vg, lv, acceptable_pvs, old_le_count,
-		       stripes, stripe_size)) {
+	if (!_allocate(lv->vg, lv, allocatable_pvs, old_le_count,
+		       stripes, stripe_size, NULL, 0u, 0u)) {
 		lv->le_count = old_le_count;
 		lv->size = old_size;
+		stack;
 		return 0;
 	}
 
@@ -519,6 +655,35 @@ int lv_extend(struct format_instance *fi,
 	return 1;
 }
 
+int lv_extend_mirror(struct format_instance *fid,
+		     struct logical_volume *lv,
+		     struct physical_volume *mirrored_pv,
+		     uint32_t mirrored_pe,
+		     uint32_t extents, struct list *allocatable_pvs,
+		     uint32_t status)
+{
+	uint32_t old_le_count = lv->le_count;
+	uint64_t old_size = lv->size;
+
+	lv->le_count += extents;
+	lv->size += (uint64_t) extents *lv->vg->extent_size;
+
+	if (!_allocate(lv->vg, lv, allocatable_pvs, old_le_count,
+		       1, extents, mirrored_pv, mirrored_pe, status)) {
+		lv->le_count = old_le_count;
+		lv->size = old_size;
+		stack;
+		return 0;
+	}
+
+	if (fid->fmt->ops->lv_setup && !fid->fmt->ops->lv_setup(fid, lv)) {
+		stack;
+		return 0;
+	}
+
+	return 1;
+}
+
 int lv_remove(struct volume_group *vg, struct logical_volume *lv)
 {
 	struct list *segh;
@@ -541,3 +706,36 @@ int lv_remove(struct volume_group *vg, struct logical_volume *lv)
 
 	return 1;
 }
+
+/* Unlock list of LVs */
+int unlock_lvs(struct cmd_context *cmd, struct list *lvs)
+{
+	struct list *lvh;
+	struct logical_volume *lv;
+
+	list_iterate(lvh, lvs) {
+		lv = list_item(lvh, struct lv_list)->lv;
+		unlock_lv(cmd, lv->lvid.s);
+	}
+
+	return 1;
+}
+
+/* Lock a list of LVs */
+int lock_lvs(struct cmd_context *cmd, struct list *lvs, int flags)
+{
+	struct list *lvh;
+	struct logical_volume *lv;
+
+	list_iterate(lvh, lvs) {
+		lv = list_item(lvh, struct lv_list)->lv;
+		if (!lock_vol(cmd, lv->lvid.s, flags)) {
+			log_error("Failed to lock %s", lv->name);
+			/* FIXME Only unlock the locked ones */
+			unlock_lvs(cmd, lvs);
+			return 0;
+		}
+	}
+
+	return 1;
+}
