@@ -8,7 +8,7 @@
 #include "metadata.h"
 #include "toolcontext.h"
 
-int lv_is_origin(struct logical_volume *lv)
+int lv_is_origin(const struct logical_volume *lv)
 {
 	struct list *slh;
 	struct snapshot *s;
@@ -22,7 +22,7 @@ int lv_is_origin(struct logical_volume *lv)
 	return 0;
 }
 
-int lv_is_cow(struct logical_volume *lv)
+int lv_is_cow(const struct logical_volume *lv)
 {
 	struct list *slh;
 	struct snapshot *s;
@@ -36,7 +36,7 @@ int lv_is_cow(struct logical_volume *lv)
 	return 0;
 }
 
-struct snapshot *find_origin(struct logical_volume *lv)
+struct snapshot *find_origin(const struct logical_volume *lv)
 {
 	struct list *slh;
 	struct snapshot *s;
@@ -50,7 +50,7 @@ struct snapshot *find_origin(struct logical_volume *lv)
 	return NULL;
 }
 
-struct snapshot *find_cow(struct logical_volume *lv)
+struct snapshot *find_cow(const struct logical_volume *lv)
 {
 	struct list *slh;
 	struct snapshot *s;
@@ -64,7 +64,7 @@ struct snapshot *find_cow(struct logical_volume *lv)
 	return NULL;
 }
 
-struct list *find_snapshots(struct logical_volume *lv)
+struct list *find_snapshots(const struct logical_volume *lv)
 {
 	struct list *slh;
 	struct list *snaplist;
@@ -137,6 +137,7 @@ int vg_add_snapshot(struct logical_volume *origin,
 	cow->status &= ~VISIBLE_LV;
 	sl->snapshot = s;
 	list_add(&origin->vg->snapshots, &sl->list);
+	origin->vg->snapshot_count++;
 
 	return 1;
 }
@@ -151,6 +152,7 @@ int vg_remove_snapshot(struct volume_group *vg, struct logical_volume *cow)
 
 		if (sl->snapshot->cow == cow) {
 			list_del(slh);
+			vg->snapshot_count--;
 			return 1;
 		}
 	}
